// pecco -- please enjoy classification with conjunctive features
//  $Id: llm.cpp 474 2010-08-13 16:17:00Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include "llm.h"

namespace pecco {
  
  bool LLM::_setFtrie () {
    const std::string mfn (_opt["-m"]);
  
    const std::string edafn (mfn + ".ne" + ny::DA_SUFFIX); // without alphas
    const std::string alfn  (mfn + ".alpha");

    if (_verbosity > 0) std::cerr << "Loading feature weight DA..";
    if (newer (edafn.c_str (),mfn.c_str ()) && _pke.open (edafn.c_str ()) == 0) {
      // read computed score
      FILE* reader = std::fopen (alfn.c_str (), "rb");
      if (! reader) ny::print_err (HERE "no such file: %s\n", alfn.c_str ());
      if (std::fseek (reader, 0, SEEK_END) != 0) return -1;
      ny::uint uniq = std::ftell (reader) / (_nl * sizeof (ny::fl_t));
      if (std::fseek (reader, 0, SEEK_SET) != 0) return -1;
      _feat_weight = new ny::fl_t [_nl*uniq]; // one-dimention array is faster
      std::fread (&_feat_weight[0], sizeof (ny::fl_t), uniq * _nl, reader);
      std::fclose (reader);
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    } else {
      if (_verbosity > 0) std::cerr << "not found." << std::endl;
      std::vector <FeatKey*> pke_key;
      _feat_weight = new ny::fl_t [_nl*_nunit];

      ny::uchar * s  = new ny::uchar[KEY_SIZE * _d];
      // should remember feature combination including rare feature
      for (size_t i = 0; i < _nunit; ++i) {
        ny::fv_t &f = _fc[i];
        std::sort (f.begin (), f.end ()); // don't comment out
        // compressed key
        ny::uint l = 0;
        for (ny::fv_t::reverse_iterator it = f.rbegin (); // rare to common
             it != f.rend (); ++it) {
          ny::uint fi = *it;
          ny::uint n = 0;
          _vbEncode (fi, s + l, n); l += n;
        }
        pke_key.push_back (new FeatKey (s, &_alpha_tmp_ary[i][0], l, _nl));
      }
      delete [] s;
      
      std::vector <const char*> str;
      std::vector <size_t> len;
      std::vector <int> val;

      FeatKeypLess  featkeypless;
      std::sort (pke_key.begin (), pke_key.end (), featkeypless);
      std::vector <FeatKey*>::const_iterator dit = pke_key.begin ();
      ny::uint uniq = 0;

      // uniq idential weights
      // (minimize the double array with a slight cahce-loss increase)
      std::map <std::vector <ny::fl_t>, ny::uint> w2id;
      for (; dit != pke_key.end (); ++dit) {
        str.push_back (reinterpret_cast <char*> ((*dit)->key));
        len.push_back ((*dit)->len);
        ny::fl_t* wv = (*dit)->cont;
        std::vector <ny::fl_t> w (&wv[0], &wv[_nl]);
        ny::uint id = 0;
        std::map <std::vector <ny::fl_t>, ny::uint>::const_iterator it = w2id.find (w);
        if (it == w2id.end ()) {
          for (ny::uint li = 0; li < _nl; ++li)
            _feat_weight[uniq * _nl + li] = wv[li]; // put weights for all labels
          id = uniq *_nl;
          w2id.insert (std::make_pair (w, id));
          ++uniq;
        } else {
          id = (*it).second;
        }
        val.push_back (id);
      }
      FILE* writer = std::fopen (alfn.c_str (), "wb");
      std::fwrite (&_feat_weight[0], sizeof (ny::fl_t), uniq *_nl, writer);
      std::fclose (writer);
      buildDA (&_pke, "feature weight", edafn, str, len, val, _verbosity> 0);
      for (dit = pke_key.begin (); dit != pke_key.end (); ++dit) delete *dit;
    }
    return true;
  }

  bool LLM::load () {
    
#ifdef USE_TIMER
    _model_t->startTimer ();
#endif
    const std::string mfn  (_opt["-m"]);   // model
    const std::string bmfn  (mfn + ".bin"); // compiled model
    if (newer (bmfn.c_str (), mfn.c_str ())) {
      if (_verbosity > 0) std::cerr << "Loading model parameters (cached)..";
      FILE* reader = std::fopen (bmfn.c_str (), "rb");
      std::fread (&_d,     sizeof (ny::uint), 1, reader);
      std::fread (&_nl,    sizeof (ny::uint), 1, reader);
      std::fread (&_nf,    sizeof (ny::uint), 1, reader);
      std::fread (&_nunit, sizeof (ny::uint), 1, reader);
      // label map
      for (ny::uint i = 0; i < _nl; ++i) {
        ny::uint size = 0;
        std::fread (&size, sizeof (ny::uint), 1, reader);
        char *p = new char[size + 1];
        std::fread (p, sizeof (char), size, reader);
        p[size] = '\0';
        _li2l.push_back (p);
        _l2li.insert (std::make_pair (p, i));
        if (std::strcmp (p, "+1") == 0) _tli = i;
      }
      // feature map
      _fi2fn.reserve (_nf+1); // _nf
      ny::uint fn = 0;
      ny::uint fn_max = 0;
      ny::uint fc = 1; _fi2fn.push_back (-1); // fi = 0;
      while (std::fread (&fn, sizeof (ny::uint), 1, reader)) {
        _fi2fn.push_back (fn);
        fn_max = std::max (fn, fn_max);
        // _fn2fi.insert (std::make_pair (fn, fc));
        ++fc;
      }
      std::fclose (reader);
      assert (_nf+1 == fc);
      _fn2fi.resize (fn_max + 1, 0);
      for (ny::uint fi = 1; fi <= _nf; ++fi) _fn2fi[_fi2fn[fi]] = fi;
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    } else {
      FILE *reader = std::fopen (mfn.c_str (), "r");
      if (_verbosity > 0) std::cerr << "Loading model parameters..";
      if (! reader) ny::print_err (HERE "no such file: %s\n", mfn.c_str ());

      _d = 0;
      std::map <ny::fv_t, ny::uint> fc2fci; // feature (combination) set
      std::vector <ny::fl_t> bv; // bias
      
      char*  line = 0;
      size_t read = 0;
      while (ny::getLine (reader, line, read)) {
        if (*line != '\0') {
          char* p = ny::skipChar (line, '\t'); // label
          if (_l2li.find (line) == _l2li.end ()) { // identify label
            char* label = new char[read];
            std::strcpy (label, line);
            if (std::strcmp (label, "+1") == 0) _tli = _nl;
            _l2li.insert (std::make_pair (label, _nl++));
            _li2l.push_back (label);
          }
          // read conjunctive feature
          ny::uint li = _l2li[line]; // bug if all alpha null
          char* cfstr = p;
          p = ny::skipChar (cfstr, '\t');
          ny::fv_t fnv;
          _convertCfstr2Cf (cfstr, fnv, true);
          _d = std::max (_d, static_cast <ny::uint> (fnv.size ())); // degree
          ny::uint uid = _nunit;
          std::map <ny::fv_t, ny::uint>::iterator it =
            fc2fci.lower_bound (fnv);
          if (it == fc2fci.end () || it->first > fnv) {
            fc2fci.insert (it,
                           std::map <ny::fv_t, ny::uint>::value_type (fnv, uid));
            _alpha_tmp_ary.push_back (std::vector <ny::fl_t> ());
            _fc.push_back (fnv);
            ++_nunit;
          } else {
            uid = it->second;
          }
          // read alpha
          const ny::fl_t alpha = ny::strton <ny::fl_t> (p);
          // check alpha == 0 or not
          if (*p == '-') ++p;
          if (*p++ == '0' && *p++ == '.') // "0.XX"
            while (*p != '\0' && *p == '0') ++p;
          if (*p == '\0') continue;
          while (_alpha_tmp_ary[uid].size () <= li)
            _alpha_tmp_ary[uid].push_back (0.0);
          _alpha_tmp_ary[uid][li] = alpha;
        }
      }
      std::fclose (reader);
      for (size_t i = 0; i < _nunit; ++i)
        while (_alpha_tmp_ary[i].size () < _nl)
          _alpha_tmp_ary[i].push_back (0.0);
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    
      // construct dense feature;
      _nf = static_cast <ny::uint> (_fncnt.size ());
      _renameFeatureID (_fc); // need afn

      // output map
      FILE* writer = std::fopen (bmfn.c_str (), "wb");
      std::fwrite (&_d,     sizeof (ny::uint), 1, writer);
      std::fwrite (&_nl,    sizeof (ny::uint), 1, writer);
      std::fwrite (&_nf,    sizeof (ny::uint), 1, writer);
      std::fwrite (&_nunit, sizeof (ny::uint), 1, writer);
      for (size_t i = 0; i < _nl; ++i) {
        const char* p = _li2l[i];
        ny::uint size = static_cast <ny::uint> (std::strlen (p));
        std::fwrite (&size, sizeof (ny::uint), 1,    writer);
        std::fwrite (p,     sizeof (char),     size, writer);
      }
      for (size_t fi = 1; fi <= _nf; ++fi)
        std::fwrite (&_fi2fn[fi], sizeof (ny::uint), 1, writer);
      std::fclose (writer);
      _fncnt.clear ();
    }
    // construct double array to get alpha
    _setFtrie ();
    // cache alpha
    if (_mode == FST) _setFStrie ();
#ifdef USE_TIMER
    _model_t->stopTimer ();
#endif
    if (_verbosity > 0) printParam ();
    return true;
  }
  
  bool LLM::batch () {

    if (_verbosity > 0) std::cerr << "Classifying events..";

    // classify one example (feature vector)
    const std::string tfn (_opt["-tf"]);
    FILE* reader = (tfn == "-") ? stdin : std::fopen (tfn.c_str (), "r");
    if (! reader) ny::print_err (HERE "no such file: %s\n", tfn.c_str ());
    
    char * line = 0;
    size_t read = 0;
    ny::uint pp (0), np (0);
    ny::fv_t fnv;
    ny::fl_t * score = new ny::fl_t[_nl];
    
    while (ny::getLine (reader, line, read)) {
      if (*line != '\0') {
        // char* p = ny::skipChar (line, ' '); // label
        const char* p_end = line + read - 1;
        char* p = line; while (p != p_end && *p != ' ' && *p != '\t') ++p;
        *p = '\0';
        char* label = line;
        fnv.clear ();
        while (p != p_end) {
          char* fs = ++p;
          while (p != p_end && *p != ':') ++p; *p = '\0';
          while (p != p_end && *p != ' ' && *p != '\t') ++p; *p = '\0';
          fnv.push_back (static_cast <ny::uint> (std::strtol (fs, NULL, 10)));
        }
        classify (fnv, score);
        ny::uint max_li = std::max_element (&score[0], &score[_nl]) - &score[0];
        char* label_ = _li2l[max_li];
        if (_verbosity > 1)
          std::fprintf (stdout, "%s %f\n", label_, _calcProb (score, max_li));
        if (std::strcmp (label_, label) == 0) ++pp; else ++np;
      }
    }
    delete [] score;
    if (reader != stdin) std::fclose (reader);
    if (_verbosity > 0)  std::cerr << "done." << std::endl;
    std::fprintf (stderr, "acc. %2.3f%% (%d/%d)\n",
                  pp * 100.0 / (pp + np), pp, pp + np);
#ifdef USE_PROFILE
    if (_verbosity > 0) printHitRatio ();
#endif
#ifdef USE_TIMER
    if (_verbosity > 0) printTimer ();
#endif
    // parameter setting
    return true;
  }
}
