// pecco -- please enjoy classification with conjunctive features
//  $Id: opt_parse.h 472 2010-08-13 13:43:20Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef OPTION_PARSE_H
#define OPTION_PARSE_H

#include <iostream>
#include <string>
#include <algorithm>
#include <map>
#include "typedef.h"

namespace ny {
  
  class Options {
  private:
    // type alias
    typedef std::map <std::string, std::string> opt_t;
    typedef std::vector <std::pair <std::string, std::string> > label_t;
    // variables
    const std::string _prog;
    opt_t             _opt;
    label_t           _lv;
    // private function
    static void _print (const std::pair <std::string, std::string> &kv)
    { std::cerr << "[" << kv.first << "=" << kv.second << "]"; }
  public:
    Options (const char* prog) : _prog (prog), _opt (), _lv () {};
  
    void set (const std::string &key, const std::string &val,
              const std::string label = std::string ("")) {
      std::pair <opt_t::iterator, bool> p
        = _opt.insert (opt_t::value_type (key, val));
      if (p.second && label != "") // register
        _lv.push_back (label_t::value_type (key, label));
      else          // overwrite
        p.first->second = val;
    };
    const std::string& operator[] (const std::string &key)  {
      opt_t::iterator it = _opt.find (key);
      if (it == _opt.end ())
        ny::print_err (HERE "unregistered option: %s\n", key.c_str ());
      return it->second;
    };
    void print () {
      std::cerr << "Options: ";
      for_each (_opt.begin (), _opt.end (), &_print);
      std::cerr << std::endl;
    };
    void printHelp () {
      std::cerr << _prog << std::endl;;
      std::cerr << "Options:" << std::endl;
      label_t::iterator it;
      for (it = _lv.begin (); it != _lv.end (); ++it)
        std::cerr << "\t" << it->first << "\t" << it->second
                  << " (" << _opt[it->first] << ")" << std::endl;
      std::cerr << std::endl;
      std::exit (1);
    };
    void parse (const std::vector <std::string>& argvec) {
      for (size_t i = 1; i < argvec.size (); ++i) {
        if (_opt.find (argvec[i]) == _opt.end ()) {
          if (argvec[i] != "-h")
            ny::print_err ("unknown option: %s\n", argvec[i].c_str ());
          printHelp ();
        } else {
          set (argvec[i], argvec[i+1]); ++i;
        }
      }
    };
  };
}

#endif /* OPTION_PARSE_H */
