// pecco -- please enjoy classification with conjunctive features
//  $Id: polyk.cpp 503 2011-01-10 07:08:56Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include "polyk.h"

namespace pecco {
  
  bool PolyK::_splitInit () {
    if (_d == 1 && _mode == FST)
      std::cerr << "NOTE: _mode = fst is useless for d = 1." << std::endl;
    ny::fl_t ec (0), ec_c (0), ec_r (0); 
    for (ny::uint fi = 1; fi <= _nf; ++fi) {
      const ny::uint fn    = _fi2fn[fi];
      const ny::fl_t &size = static_cast <ny::fl_t> (_fncnt[fn]);
#ifdef USE_FREQ_SORT
      ny::fl_t ratio = size / static_cast <ny::fl_t> (_nt);
#elif defined(USE_UNIT_SORT)
      ny::fl_t ratio = size / static_cast <ny::fl_t> (_nunit);
#endif
      ratio *= static_cast <ny::fl_t> (100.0);
      if (ratio >= _feat_cutoff_ratio) ec_c += size; else ec_r += size; 
      ec += size;
    }
    return true;
  }
  bool PolyK::_precomputeKernel () {
    if (_verbosity > 0) std::cerr << "Precomputing kernel cache..";
    _polyk = new ny::fl_t[_maf+1];
    if (_mode == SPLIT || _mode == FST) _spolyk = new ny::fl_t[_maf+1];
    for (ny::uint i = 0; i <= _maf; ++i) { // bug; i -> dot product rare
      _polyk[i] = static_cast <ny::fl_t>
                  (std::pow (_s * static_cast <ny::fl_t> (i) + _r, static_cast <int> (_d)));
      if (_mode == SPLIT || _mode == FST)
        _spolyk[i] = powf (static_cast <float> (_s + _s * i + _r), _d)
                     - powf (static_cast <float> (_s * i + _r), _d);
    }
    if (_verbosity > 0) std::cerr << "done." << std::endl;
    return true;
  }
  bool PolyK::load () {
#ifdef USE_TIMER
    _model_t->startTimer ();
#endif
    const std::string mfn  (_opt["-m"]);
    const std::string bmfn (mfn + ".bin");

    if (newer (bmfn.c_str (), mfn.c_str ())) {
      if (_verbosity > 0) std::cerr << "Loading model parameters (cached)..";
      FILE* reader = std::fopen (bmfn.c_str (), "rb");
      // model parameters
      std::fread (&_d,     sizeof (ny::uint), 1, reader);
      std::fread (&_b,     sizeof (ny::fl_t), 1, reader);
      std::fread (&_r,     sizeof (ny::fl_t), 1, reader);
      std::fread (&_s,     sizeof (ny::fl_t), 1, reader);
      std::fread (&_maf,   sizeof (ny::uint), 1, reader);
      std::fread (&_nf,    sizeof (ny::uint), 1, reader);
      std::fread (&_nunit, sizeof (ny::uint), 1, reader);
      // feature map
      _fi2fn.reserve (_nf+1); // _nf
      ny::uint fn = 0;
      ny::uint fn_max = 0;
      ny::uint fc = 1; _fi2fn.push_back (-1); // fi = 0;
      ny::fl_t ratio = 0;
      while (std::fread (&fn,    sizeof (ny::uint), 1, reader) &&
             std::fread (&ratio, sizeof (ny::fl_t), 1, reader)) {
        if (ratio >= _feat_cutoff_ratio) _f_r = fc + 1;
        _fi2fn.push_back (fn);
        fn_max = std::max (fn, fn_max);
        //_fn2fi.insert (std::make_pair (fn, fc));
        ++fc;
        if (fc > _nf) break; // fi >= _nf
      }
      _fn2fi.resize (fn_max + 1, 0);
      for (ny::uint fi = 1; fi <= _nf; ++fi) _fn2fi[_fi2fn[fi]] = fi;
      if (_mode != PKI) _sv.resize (_nunit, sv_t (ny::fv_t (), 0));
      if (_mode == PKI) _alph.reserve (_nunit);
      if (_mode == PKI || _mode == SPLIT || _mode == FST)
        _f2ss.resize (_nf + 1); // _nf
      ny::fl_t alpha = 0;
      ny::uint size  = 0;
      ny::uint i     = 0;
      ny::uint fi    = 0;
      while (std::fread (&alpha, sizeof (ny::fl_t), 1, reader) &&
             std::fread (&size,  sizeof (ny::uint), 1, reader)) {
        if (_mode == PKI) {
          _alph.push_back (alpha);
          for (size_t j = 0; j < size; ++j) {
            std::fread (&fi, sizeof (ny::uint), 1, reader);
            _f2ss[fi].push_back (i);
          }
        } else {
          _sv[i].alpha = alpha;
          _sv[i].fv.reserve (size);
          for (size_t j = 0; j < size; ++j) {
            std::fread (&fi, sizeof (ny::uint), 1, reader);
            _sv[i].fv.push_back (fi);
            if (fi >= _f_r && (_mode == SPLIT || _mode == FST))
              _f2ss[fi].push_back (i);
          }
        }
        if (++i >= _nunit) break;
      }
      std::fclose (reader);
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    } else {
      if (_verbosity > 0) std::cerr << "Loading model parameters..";
      FILE* reader = std::fopen (mfn.c_str (), "r");
      if (! reader) ny::print_err (HERE "No such file: %s\n", mfn.c_str ());
      bool   header = true;
      char*  line   = 0;
      size_t read   = 0;
      std::vector <ny::fv_t> unit;
      while (ny::getLine (reader , line, read)) { // parse
        if (*line != '\0') {
          if (! header) {
            char* p = ny::skipChar (line, ' ');
            _alph.push_back (ny::strton <ny::fl_t> (line));
            ny::fv_t fnv;
            _convertEx2Fv (p, fnv, true);
            _maf = std::max (_maf, static_cast <ny::uint> (fnv.size ()));
            unit.push_back (fnv);
            ++_nunit;
          } else if (std::strstr (line, "kernel type") != NULL) {
            if (*line != '1') ny::print_err (HERE "unsupported kernel used.\n");
          } else if (std::strstr (line, "kernel") != NULL &&
                     std::strstr (line, "empty")  == NULL) {
            if (std::strchr (line, '-') == NULL) continue;
            char* p = ny::skipChar (line, ' ');
            switch (*ny::skipChar (p, '-')) {
              case 'r': _r = ny::strton <ny::fl_t> (line); break;
              case 's': _s = ny::strton <ny::fl_t> (line); break;
              case 'd': _d = ny::strton <ny::uint> (line); break;
            }
          } else if (std::strstr (line, " # threshold b") != NULL) {
            _b = ny::strton <ny::fl_t> (std::strtok (line, " "));
            header = false;
          }
        }
      }
      std::fclose (reader);
      if (_verbosity > 0) std::cerr << "done." << std::endl;
      // Zipf-aware feature indexing
      _nf = static_cast <ny::uint> (_fncnt.size ());
      _renameFeatureID (unit);
      // report splitSVM status
      if (_mode == SPLIT || _mode == FST) _splitInit ();
      // output map
      FILE* writer = std::fopen (bmfn.c_str (), "wb");
      std::fwrite (&_d,     sizeof (ny::uint), 1, writer);
      std::fwrite (&_b,     sizeof (ny::fl_t), 1, writer);
      std::fwrite (&_r,     sizeof (ny::fl_t), 1, writer);
      std::fwrite (&_s,     sizeof (ny::fl_t), 1, writer);
      std::fwrite (&_maf,   sizeof (ny::uint), 1, writer);
      std::fwrite (&_nf,    sizeof (ny::uint), 1, writer);
      std::fwrite (&_nunit, sizeof (ny::uint), 1, writer);
      // common-to-rare ranking of features
      for (ny::uint fi = 1; fi <= _nf; ++fi) { // fi = 0; fi < _nf
        ny::uint fn = _fi2fn[fi];
        ny::fl_t ratio = static_cast <ny::fl_t> (_fncnt[fn]);
#ifdef USE_FREQ_SORT
        ratio /= static_cast <ny::fl_t> (_nt);
#elif defined(USE_UNIT_SORT)
        ratio /= static_cast <ny::fl_t> (_nunit);
#endif
        ratio *= static_cast <ny::fl_t> (100.0);
        if (ratio >= _feat_cutoff_ratio) _f_r = fi + 1;
        std::fwrite (&fn,    sizeof (ny::uint), 1, writer);
        std::fwrite (&ratio, sizeof (ny::fl_t), 1, writer);
      }
      // support vectors
      _f2ss.resize (_nf+1, ss_t ());
      for (size_t i = 0; i < _nunit; ++i) {
        const ny::fl_t w   = _alph[i];
        const ny::fv_t &fv = unit[i];
        ny::uint len = static_cast <ny::uint> (fv.size ());
        std::fwrite (&w,     sizeof (ny::fl_t), 1,   writer);
        std::fwrite (&len,   sizeof (ny::uint), 1,   writer);
        std::fwrite (&fv[0], sizeof (ny::uint), len, writer);
        if (_mode != PKI) _sv.push_back (sv_t (fv, w));
        if (_mode == PKI || _mode == SPLIT || _mode == FST)
          for (ny::fv_it it = fv.begin (); it != fv.end (); ++it) {
            if (_mode != PKI && *it < _f_r) continue;
            _f2ss[*it].push_back (i);
          }
      }
      if (_mode == PKI || _mode == SPLIT || _mode == FST)
        std::vector <ny::fl_t> (_alph).swap (_alph); // shrink
      else
        std::vector <ny::fl_t> ().swap (_alph);      // clear
      _fncnt.clear ();
      std::fclose (writer);
    }
    _fv.reserve (_maf);
    // calculate dot product
    if (_mode != PKE ) _precomputeKernel ();
    // used to indicate active features
    if (_mode == PKI)
      _dot = new ny::uint[_nunit];
    if (_mode != PKB && _mode != PKI) {
      _fbit = new bool[_nf+1]; // _nf
      std::fill (&_fbit[0], &_fbit[_nf+1], false); // &_fbit[1]
    }
    // kernel expansion
    _nf_cut = _nf;
    if (_mode == PKE || _mode == SPLIT || _mode == FST) _setFtrie ();
    if (_sigma > static_cast <ny::fl_t> (0) || _f_r-1 < _nf) // _f_r != _nf
      if (_mode == PKE ||
          ((_mode == SPLIT || _mode == FST) && _sigma > 0.0) ||
          ((_mode == PKB || _mode == PKBPP) && _f_r - 1 < _nf))
        std::cerr << "NOTE (approximated computation); "
                  << _f_r-1 << "/" << _nf << " features used, "
                  << "PKE sigma=" << _sigma << std::endl;
    // fstrie construction
    if (_mode == FST) _setFStrie ();
#ifdef USE_TIMER
    _model_t->stopTimer ();
#endif
    if (_verbosity > 0) printParam ();
    if (_mode == PKBPP || _mode == SPLIT || _mode == FST)
      std::fill (&_fbit[1], &_fbit[_nf+1], false);
    return true;
  }
  bool PolyK::_setPKEcoeff () { // coefficients for conjunctive features
    switch (_d) {
      case 1:
        _coeff[0] = _r;
        _coeff[1] = _s; break;
      case 2:
        _coeff[0] = _r*_r;
        _coeff[1] = _s*_s+2*_r*_s;
        _coeff[2] =  2*_s*_s; break;
      case 3:
        _coeff[0] = _r*_r*_r;
        _coeff[1] = _s*_s*_s + 3*_r*_s*_s + 3*_r*_r*_s;
        _coeff[2] = 6*_s*_s*_s + 6*_r*_s*_s;
        _coeff[3] = 6*_s*_s*_s; break;
      case 4:
        _coeff[0] = _r*_r*_r*_r;
        _coeff[1] = _s*_s*_s*_s + 4*_r*_s*_s*_s + 6*_r*_r*_s*_s + 4*_r*_r*_r*_s;
        _coeff[2] = 14*_s*_s*_s*_s + 24*_r*_s*_s*_s  + 12*_r*_r*_s*_s;
        _coeff[3] = 36*_s*_s*_s*_s + 24*_r*_s*_s*_s;
        _coeff[4] = 24*_s*_s*_s*_s; break;
      default: assert (! "please add case statement.");
    }
    _max_coeff = *std::max_element (&_coeff[0], &_coeff[_d+1]);
    return true;
  }
  // implementation of kernel expansion (Kudo & Matsumoto, 2003)
  void PolyK::_pkePrefixSpan (ny::fv_t &fc,
                              const std::vector <std::pair <ny::uint, int> > &proj,
                              std::vector <FeatKey*> &pke_key) {
    // pseudo projection to mine conjunctive features
    typedef std::vector <std::pair <ny::uint, int> > proj_t;
    typedef ny::map <ny::uint, proj_t>::type proj_map_t;
    proj_map_t proj_map;
    for (proj_t::const_iterator pit = proj.begin ();
         pit != proj.end (); ++pit) {
      const ny::uint i   = pit->first;
      const ny::fv_t &fv = _sv[i].fv;
      for (int j = pit->second; j >= 0; --j)
        proj_map[fv[j]].push_back (std::make_pair (i, j-1));
    }
    ny::uchar * s  = new ny::uchar[KEY_SIZE * _d];
    for (proj_map_t::const_iterator pmit = proj_map.begin ();
         pmit != proj_map.end (); ++pmit) {
      // extend span
      const ny::uint fi        = pmit->first;
      const proj_t &proj_child = pmit->second;
      ny::fl_t mu_pos (0), mu_neg (0), w (0);
      fc.push_back (fi);
      bool terminate = (fc.size () >= _d || proj_child.empty ());
      for (proj_t::const_iterator pit = proj_child.begin ();
           pit != proj_child.end (); ++pit) {
        const ny::fl_t alpha = _sv[pit->first].alpha;
        w += static_cast <ny::fl_t> (_coeff[fc.size ()]) * alpha;
        if (! terminate) {
          if (alpha > 0)
            mu_pos += static_cast <ny::fl_t> (_max_coeff) * alpha;
          else
            mu_neg += static_cast <ny::fl_t> (_max_coeff) * alpha;
        }
      }
      terminate |= (mu_neg > _sigma_neg && mu_pos < _sigma_pos); // pruning
      if (w <= _sigma_neg || w >= _sigma_pos) { // approximation
        ny::uint l = 0;
        for (ny::fv_t::iterator it = fc.begin ();
             it != fc.end (); ++it) {
          const ny::uint fi = *it;
          _fbit[fi] = true;
          ny::uint n = 0; _vbEncode (fi, s + l, n); l += n;
        }
        pke_key.push_back (new FeatKey (s, &w, l, _nl));
      }
      if (!terminate) _pkePrefixSpan (fc, proj_child, pke_key);
      if (fc.size () == 1)
        if (_verbosity > 0)
          std::cerr << "\r " << ++_processed << "/" << _f_r-1 << " feats. "
                    << "processed => " << pke_key.size () << " keys generated..";
      fc.pop_back ();
    }
    delete [] s;
  }
  // polynomial kernel expanded
  bool PolyK::_setFtrie () {
    const std::string mfn (_opt["-m"]);

    std::ostringstream ss;
    ss << mfn << ".r" << _opt["-r"] << ".s" <<  _opt["-s"];
#ifdef ABUSE_DA
    const std::string edafn (ss.str () + ".e"  + ny::DA_SUFFIX);
#else
    const std::string edafn (ss.str () + ".ne" + ny::DA_SUFFIX); // w/o alphas
    const std::string exalfn (ss.str () + ".alpha");
#endif
    const std::string affn (ss.str () + ".af");
    // set PKE coeff
    _setPKEcoeff (); // calculte bias w(0) !!! bug
    if (_verbosity > 0) std::cerr << "Loading feature weight DA..";
    for (size_t i = 0; i < _nunit; ++i) _b -= _coeff[0] * _sv[i].alpha;
    if (newer (edafn.c_str (), mfn.c_str ()) && _pke.open (edafn.c_str ()) == 0) {
      // read computed score
      FILE* reader = std::fopen (affn.c_str (), "rb");
      if (! reader) ny::print_err (HERE "no such file: %s\n", affn.c_str ());
      std::fread (&_exnunit, sizeof (ny::uint), 1,    reader);
      std::fread (&_fbit[0], sizeof (bool),    _nf+1, reader); // _nf
      std::fclose (reader);
#ifndef ABUSE_DA
      reader = std::fopen (exalfn.c_str (), "rb");
      if (! reader) ny::print_err (HERE "no such file: %s\n", exalfn.c_str ());
      if (std::fseek (reader, 0, SEEK_END) != 0) return -1;
      ny::uint uniq = std::ftell (reader) / sizeof (ny::fl_t);
      if (std::fseek (reader, 0, SEEK_SET) != 0) return -1;
      _feat_weight = new ny::fl_t [uniq];
      std::fread (&_feat_weight[0], sizeof (ny::fl_t), uniq, reader);
      std::fclose (reader);
#endif
      if (_verbosity > 0) std::cerr << "done." << std::endl;
      _exnunit_cut = _exnunit;
    } else {
      if (_verbosity > 0) std::cerr << "not found." << std::endl;
      if (_verbosity > 0) std::cerr << "Calculating feature weight by PKE..";
      if (_f_r == 0) { // no common feature
        if (_verbosity > 0) std::cerr << "skipped" << std::endl;
        return false;
      }
      if (_verbosity > 0) std::cerr << std::endl;
      size_t pos_num = 0;
      size_t neg_num = 0;
      // pseudo projection (PrefixSpan)
      std::vector <std::pair <ny::uint, int> > proj;
      for (size_t i = 0; i < _nunit; ++i) {
        const ny::fv_t &fv = _sv[i].fv;
        int tail = std::distance (fv.begin (), _splitFv (fv)) - 1;
        proj.push_back (std::make_pair (i, tail));
        if (_sv[i].alpha > 0) ++pos_num; else ++neg_num;
      }
      _sigma_pos =  _sigma * static_cast <ny::fl_t> (pos_num) / static_cast <ny::fl_t> (_nunit);
      _sigma_neg = -_sigma * static_cast <ny::fl_t> (neg_num) / static_cast <ny::fl_t> (_nunit);
      if (_verbosity > 0)
        std::cerr << " " << _f_r-1 << "/" <<_nf << " feats., "
                  << "_sigma_pos=" << _sigma_pos << ", "
                  << "_sigma_neg=" << _sigma_neg << std::endl;
      _processed = 0;
      ny::fv_t fc;
      std::vector <FeatKey*> pke_key;
      _pkePrefixSpan (fc, proj, pke_key);
      _exnunit = static_cast <ny::uint> (pke_key.size ());
      _exnunit_cut = _exnunit;
      _nf_cut = 0;
#ifndef ABUSE_DA
      _feat_weight = new ny::fl_t [_exnunit];
#endif
      for (ny::uint fi = 1; fi <= _nf; ++fi) if (_fbit[fi]) ++_nf_cut;
      if (_verbosity > 0)
        std::cerr << "\r " << _f_r-1 << " features processed => " << _exnunit
                  << " keys generated..done." << std::endl
                  << " # active features after thresholding " << _nf << " -> "
                  << _nf_cut << std::endl;
      if (_verbosity > 0) std::cerr << "done." << std::endl;
      // build feature weight DA
      std::vector <const char*> str;
      std::vector <size_t> len;
      std::vector <int> val;
      str.reserve (_exnunit);
      len.reserve (_exnunit);
      val.reserve (_exnunit);
      
      FeatKeypLess  featkeypless;
      std::sort (pke_key.begin (), pke_key.end (), featkeypless);
      std::vector <FeatKey*>::const_iterator dit = pke_key.begin ();
#ifndef ABUSE_DA
      ny::uint uniq = 0;
#endif
      typedef ny::map <ny::fl_t, ny::uint>::type w2id_t;
      w2id_t w2id;
      for (; dit != pke_key.end (); ++dit) {
        str.push_back (reinterpret_cast <char*> ((*dit)->key));
        len.push_back ((*dit)->len);
        ny::fl_t* wv = (*dit)->cont;
#ifdef ABUSE_DA
        union byte_4 b4 (static_cast <float> (*wv)); b4.u >>= 1; val.push_back (b4.i);
#else
        ny::uint uid = 0;
        // uniq identical weights
        // (minimize the double array plus a slight cache-loss increase)
        std::pair <w2id_t::iterator, bool> itb
          = w2id.insert (w2id_t::value_type (*wv, uniq));
        if (itb.second)
          { uid = uniq++; _feat_weight[uid] = *wv; }
        else
          { uid = itb.first->second; }
        val.push_back(uid);
#endif
      }
      FILE* writer = std::fopen (affn.c_str (), "wb");
      std::fwrite (&_exnunit, sizeof (ny::uint), 1,     writer);
      std::fwrite (&_fbit[0], sizeof (bool),     _nf+1, writer); // _nf
#ifndef ABUSE_DA
      writer = std::fopen (exalfn.c_str (), "wb");
      std::fwrite (&_feat_weight[0], sizeof (ny::fl_t), uniq, writer);
#endif
      std::fclose (writer);
      buildDA (&_pke, "feature weight", edafn, str, len, val, _verbosity > 0);
      for (dit = pke_key.begin (); dit != pke_key.end (); ++dit) delete *dit;
    }
    _nf_cut = 0;
    for (ny::uint fi = 1; fi <= _nf; ++fi) if (_fbit[fi]) ++_nf_cut;
    return true;
  }
  bool PolyK::batch () { // pecco interface; batch classification
    if (_verbosity > 0) std::cerr << "Classifying events..";
    const std::string tfn (_opt["-tf"]);
    FILE* reader = (tfn == "-") ? stdin : std::fopen (tfn.c_str (), "r");
    if (! reader) ny::print_err (HERE "no such file: %s\n", tfn.c_str());
    
    char*  line = 0;
    size_t read = 0;
    ny::uint pp (0), pn (0), np (0), nn (0);
    ny::fv_t fnv;
    while (ny::getLine (reader, line, read)) {
      if (*line != '\0') {
        // char* p   = ny::skipChar (line, ' '); // label
        const char* p_end = line + read - 1;
        char* p = line; while (p != p_end && *p != ' ' && *p != '\t') ++p;
        *p = '\0';
        int label = ny::strton <int> (line);
        fnv.clear ();
        while (p != p_end) {
          char* fs = ++p;
          while (p != p_end && *p != ':') ++p; *p = '\0';
          while (p != p_end && *p != ' ' && *p != '\t') ++p; *p = '\0';
          fnv.push_back (static_cast <ny::uint> (std::strtol (fs, NULL, 10)));
        }
        ny::fl_t score = classify (fnv);
        if (_verbosity > 1) std::fprintf (stdout, "%d %f\n", label, score);
        if (score >= 0) if (label > 0) ++pp; else ++np;
        else            if (label > 0) ++pn; else ++nn;
      }
    }
    if (reader != stdin) std::fclose (reader);
    if (_verbosity > 0) std::cerr << "done." << std::endl;
    std::fprintf (stderr, "acc. %2.3f%% (pp %d) (pn %d) (np %d) (nn %d)\n",
                  (pp + nn) * 100.0 / (pp + pn + np + nn), pp, pn, np, nn);
#ifdef USE_PROFILE
    if (_verbosity > 0) printHitRatio ();
#endif
    // parameter setting
#ifdef USE_TIMER
    if (_verbosity > 0) printTimer ();
#endif
    return true;
  }
#ifndef NDEBUG
  const ny::fl_t* PolyK::_pkbClassify (const ny::fv_t &fv, ny::fl_t* score) const {
    for (std::vector <sv_t>::const_iterator sit = _sv.begin ();
         sit != _sv.end (); ++sit) {
      const sv_t &s = *sit;
      ny::uint dot = 0;
      ny::fv_it i = fv.begin ();
      const ny::fv_it i_end = fv.end ();
      ny::fv_it j = s.begin ();
      const ny::fv_it j_end = s.end ();
    
      while (i != i_end && j != j_end) {
        if      (*i == *j) { ++dot; ++i; ++j; }
        else if (*i >  *j) ++j;
        else               ++i;
      }
      *score += s.alpha * _polyk[dot];
    }
    return score;
  }
  const ny::fl_t* PolyK::_pkbppClassify (const ny::fv_t &fv, ny::fl_t* score) {
    // fast PKB for binary features
    for (ny::fv_it fit = fv.begin (); fit != fv.end (); ++fit)
      _fbit[*fit] = true;
    for (std::vector <sv_t>::const_iterator it = _sv.begin ();
         it != _sv.end (); ++it) {
      ny::uint dot = 0;
      for (ny::fv_it sit = it->begin (); sit != it->end (); ++sit)
        if (_fbit[*sit]) ++dot;
      *score += it->alpha * _polyk[dot];
    }
    for (ny::fv_it fit = fv.begin (); fit != fv.end (); ++fit)
      _fbit[*fit] = false;
    return score;
  }
#endif
  const ny::fl_t* PolyK::_pkiClassify (const ny::fv_t &fv, ny::fl_t* score) const {
    std::fill (&_dot[0], &_dot[_nunit], 0); // slower than gcc extensions
    for (ny::fv_it it = fv.begin (); it != fv.end (); ++it) {
      const ss_t &ss = _f2ss[*it];
      for (ss_it st = ss.begin (); st != ss.end (); ++st) ++_dot[*st];
    }
    for (size_t i = 0; i < _nunit; ++i) *score += _polyk[_dot[i]] * _alph[i];
    return score;
  }
  // implementation of splitSVM (Goldburg & Elhadad, 2008)
  ny::fl_t* PolyK::_splitClassify (const ny::fv_t &fv, ny::fl_t* score, ny::fv_it cit, const ny::fv_it end) {
    ny::fv_it rit;
    if (_f_r > _nf || (rit = _splitFv (fv)) >= end) { // _f_r == _nf; for fst
      return _pkeClassify (fv, score, cit, end); // run perfect PKE
    } else { // calculate remaining common part by PKE
      if (cit < rit) {
        score = _pkeClassify (fv, score, cit, rit);
        cit = rit;
      }
    }
#ifdef USE_PROFILE
    _hit += cit - fv.begin ();
#endif
    // rare part by incremental split polynomial kernel
    for (ny::fv_it it = fv.begin (); it != cit; ++it) _fbit[*it] = true;
    for (ny::fv_it it = cit; it != end; ++it) {
      const ss_t &ss = _f2ss[*it];
      for (ss_it st = ss.begin (); st != ss.end (); ++st) {
        const sv_t &s    = _sv[*st];
        ny::uint   dot_c = 0;
        for (ny::fv_it sit = s.begin (); sit != s.end (); ++sit)
          if (_fbit[*sit]) ++dot_c;
        *score += s.alpha * _spolyk[dot_c];
      }
      _fbit[*it] = true;
    }
    for (ny::fv_it it = fv.begin (); it != end; ++it) _fbit[*it] = false;
    return score;
  }
}
